import inspect
from functools import wraps

from lumigo_core.logger import get_logger

from lumigo_tracer.auto_tag.auto_tag_event import AutoTagEvent
from lumigo_tracer.lambda_tracer.spans_container import SpansContainer, TimeoutMechanism
from lumigo_tracer.lumigo_utils import (
    config,
    is_aws_environment,
    is_kill_switch_on,
    lumigo_safe_execute,
)

CONTEXT_WRAPPED_BY_LUMIGO_KEY = "_wrapped_by_lumigo"


def _is_context_already_wrapped(*args) -> bool:  # type: ignore[no-untyped-def]
    """
    This function is here in order to validate that we didn't already wrap this lambda
        (using the sls plugin / auto instrumentation / etc.)
    """
    return len(args) >= 2 and hasattr(args[1], CONTEXT_WRAPPED_BY_LUMIGO_KEY)


def _add_wrap_flag_to_context(*args):  # type: ignore[no-untyped-def]
    """
    This function is here in order to validate that we didn't already wrap this invocation
        (using the sls plugin / auto instrumentation / etc.).
    We are adding lumigo's flag to the context, and check it's value in _is_context_already_wrapped.
    """
    if len(args) >= 2:
        with lumigo_safe_execute("wrap context"):
            setattr(args[1], CONTEXT_WRAPPED_BY_LUMIGO_KEY, True)


def _lumigo_tracer(func):  # type: ignore[no-untyped-def]
    if is_kill_switch_on():
        return func

    @wraps(func)
    def lambda_wrapper(*args, **kwargs):  # type: ignore[no-untyped-def]
        if _is_context_already_wrapped(*args):
            return func(*args, **kwargs)
        _add_wrap_flag_to_context(*args)
        executed = False
        ret_val = None
        try:
            SpansContainer.create_span(*args, is_new_invocation=True)  # type: ignore[misc]
            with lumigo_safe_execute("auto tag"):
                AutoTagEvent.auto_tag_event(args[0])
            SpansContainer.get_span().start(*args)
            try:
                executed = True
                ret_val = func(*args, **kwargs)
            except Exception as e:
                with lumigo_safe_execute("Customer's exception"):
                    SpansContainer.get_span().add_exception_event(e, inspect.trace())
                raise
            finally:
                with lumigo_safe_execute("end"):
                    SpansContainer.get_span().end(ret_val, *args)
            return ret_val
        except Exception:
            # The case where our wrapping raised an exception
            if not executed:
                TimeoutMechanism.stop()
                get_logger().exception("exception in the wrapper", exc_info=True)
                return func(*args, **kwargs)
            else:
                raise

    return lambda_wrapper


def _add_prefix_for_each_line(prefix: str, text: str):  # type: ignore[no-untyped-def]
    enhanced_lines = []
    for line in text.split("\n"):
        if line and not line.startswith(prefix):
            line = prefix + " " + line
        enhanced_lines.append(line)
    return "\n".join(enhanced_lines)


def lumigo_tracer(*args, **kwargs):  # type: ignore[no-untyped-def]
    """
    This function should be used as wrapper to your lambda function.
    It will trace your HTTP calls and send it to our backend, which will help you understand it better.

    If the kill switch is activated (env variable `LUMIGO_SWITCH_OFF` set to 1), this function does nothing.

    You can pass to this decorator more configurations to configure the interface to lumigo,
        See `lumigo_tracer.reporter.config` for more details on the available configuration.
    """
    config(*args, **kwargs)
    return _lumigo_tracer


class LumigoChalice:
    DECORATORS_OF_NEW_HANDLERS = [
        "on_s3_event",
        "on_sns_message",
        "on_sqs_message",
        "schedule",
        "authorizer",
        "lambda_function",
        "on_cw_event",
        "on_dynamodb_record",
    ]

    def __new__(cls, app, *args, **kwargs):  # type: ignore[no-untyped-def]
        if is_aws_environment() and not is_kill_switch_on():
            return super().__new__(cls)
        get_logger().debug("Disabling LumigoChalice")
        return app

    def __init__(self, app, *args, **kwargs):  # type: ignore[no-untyped-def]
        self.lumigo_conf_args = args
        self.lumigo_conf_kwargs = kwargs
        self.app = app
        self.original_app_attr_getter = app.__getattribute__
        self.lumigo_app = lumigo_tracer(*self.lumigo_conf_args, **self.lumigo_conf_kwargs)(app)

    def __getattr__(self, item):  # type: ignore[no-untyped-def]
        original_attr = self.original_app_attr_getter(item)
        if is_aws_environment() and item in self.DECORATORS_OF_NEW_HANDLERS:

            def get_decorator(*args, **kwargs):  # type: ignore[no-untyped-def]
                # calling the annotation, example `app.authorizer(THIS)`
                chalice_actual_decorator = original_attr(*args, **kwargs)

                def wrapper2(func):  # type: ignore[no-untyped-def]
                    user_func_wrapped_by_chalice = chalice_actual_decorator(func)
                    return LumigoChalice(
                        user_func_wrapped_by_chalice,
                        *self.lumigo_conf_args,
                        **self.lumigo_conf_kwargs,
                    )

                return wrapper2

            return get_decorator
        return original_attr

    def __call__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        if len(args) < 2 and "context" not in kwargs:
            kwargs["context"] = getattr(getattr(self.app, "current_request", None), "context", None)
        return self.lumigo_app(*args, **kwargs)
